// Unpack Install Creator

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <windows.h>	// for CreateDirectory

#include "zlib.h"
#include "bzlib.h"


// Type Definitions for short types
typedef unsigned char	UINT8;
typedef signed char 	INT8;

typedef unsigned short	UINT16;
typedef signed short	INT16;

typedef unsigned int	UINT32;
typedef signed int		INT32;


#define INLINE	__inline


/*// FILETIME from wtypes.h
typedef struct  _FILETIME
{
	DWORD dwLowDateTime;
	DWORD dwHighDateTime;
}	FILETIME;*/


typedef struct ins_crt_file_times
{
	FILETIME Time1;
	FILETIME Time2;
	FILETIME Time3;
} INS_CRT_FILE_TIMES;
typedef struct ins_crt_file_location
{
	char* Name;
	char* Folder;
	char* Link;
} INS_CRT_FILE_LOC;
typedef struct install_creator_file
{
	// this structure is based on TOC version 3 (which I reverse-engineered first)
	UINT32 EntryLen;
	UINT32 Unknown02;
	UINT32 DataPos;
	UINT32 DataSize;	// compressed Data Size
	UINT32 Unused0E;	// meant as FilePos uncompressed??
	UINT32 FileSize;	// decompressed File Size
	UINT32 ChkSum;		// maybe checksum
	UINT32 Flags1;		// Bit 3 (08) - has File Time??
	UINT32 Flags2;		// Bit 9 (200) - Main EXE??
	UINT32 Unused22;
	UINT32 Unused26;
	UINT32 FileID;
	INS_CRT_FILE_TIMES FileTimes;
	INS_CRT_FILE_LOC Loc_Main;
	INS_CRT_FILE_LOC Loc_Desktop;
	INS_CRT_FILE_LOC Loc_QLaunch;
} INS_CRT_FILE;


int main(int argc, char* argv[]);
UINT8 UnpackInstallCreator(char* File, char* Path);
UINT32 ReadTOC(const UINT32 TOCLen, const char* TOCData, INS_CRT_FILE** RetFileStruct);
INLINE UINT32 ReadString(char** String, const char* FileData);
void CreatePath(const char* FileName);
UINT8 UnpackFile(const UINT8* Data, UINT32 DataLen, char** OutBuf, UINT32* OutLen);


int main(int argc, char* argv[])
{
	char* InstallFile;
	char* ExtractPath;
	
	printf("Install Creator Extractor\n-------------------------\n");
	if (argc <= 0x02)
	{
		printf("Usage: ICExtract.exe InstCreator.exe ExtractPath\n");
		return 1;	// need a file
	}
	
	InstallFile = argv[0x01];
	ExtractPath = (char*)malloc(strlen(argv[0x02]) + 0x02);
	strcpy(ExtractPath, argv[0x02]);
	if (ExtractPath[strlen(ExtractPath) - 0x01] != '\\')
		strcat(ExtractPath, "\\");
	
	UnpackInstallCreator(InstallFile, ExtractPath);
	
#ifdef _DEBUG
	getchar();
#endif
	
	return 0;
}

UINT8 UnpackInstallCreator(char* File, char* Path)
{
	FILE* hFile;
	UINT32 FileLen;
	UINT8* FileData;
	UINT32 OutLen;
	UINT8* OutBuf;
	UINT32 CurPos;
	UINT16* Int16Pnt;
	UINT32* Int32Pnt;
	UINT32 CurFile;
	char* OutFileName;
	UINT32 DataLen;
	UINT8 DataFound;
	UINT8* PakData;
	UINT32 FileCount;
	INS_CRT_FILE* Files;
	INS_CRT_FILE* TempFile;
	UINT8 RetVal;
	
	hFile = fopen(File, "rb");
	if (hFile == NULL)
	{
		printf("Error: Can't open Installer File!\n");
		return 0x10;
	}
	
	fseek(hFile, 0x00, SEEK_END);
	FileLen = ftell(hFile);
	
	fseek(hFile, 0x00, SEEK_SET);
	FileData = (UINT8*)malloc(FileLen);
	fread(FileData, 0x01, FileLen, hFile);
	
	fclose(hFile);
	
	// Search for data
	Int16Pnt = (UINT16*)FileData;
	DataFound = 0x00;
	for (CurPos = 0x00; CurPos < FileLen; CurPos += 0x02, Int16Pnt ++)
	{
		if (*Int16Pnt == 0x7777)
		{
			Int32Pnt = (UINT32*)(&FileData[CurPos + 0x02]);
			if (*Int32Pnt == 0x48295467)
			{
				CurPos += 0x06;
				DataFound = 0x01;
				break;
			}
		}
	}
	if (! DataFound)
	{
		printf("Unable to find data! Is this really an Install Creator file??\n");
		return 0x40;
	}
	memcpy(&DataLen, &FileData[CurPos + 0x04], 0x04);
	if (! DataLen)
	{
		printf("That seems to be an Install Maker file, which isn't supported.");
		return 0xF0;
	}
	
	CreatePath(Path);
	FileCount = 0x00;
	Files = NULL;
	OutFileName = (char*)malloc(MAX_PATH + 0x10);
	do
	{
		memcpy(&CurFile, &FileData[CurPos + 0x00], 0x04);	// some sort of ID
		memcpy(&DataLen, &FileData[CurPos + 0x04], 0x04);
		CurPos += 0x08;
		
		printf("File %X\n", CurFile);
		memcpy(&OutLen, &FileData[CurPos + 0x00], 0x04);
		RetVal = UnpackFile(&FileData[CurPos + 0x05], DataLen, &OutBuf, &OutLen);
		
		if (! RetVal)
		{
			sprintf(OutFileName, "%s%08X", Path, CurFile);
			switch(FileData[CurPos + 0x04])
			{
			case 0x01:
			case 0x02:
				strcat(OutFileName, ".bin");
				break;
			default:
				strcat(OutFileName, ".cmp");
				break;
			}
			hFile = fopen(OutFileName, "wb");
			if (hFile == NULL)
			{
				printf("Error: Can't open Output File!\n");
				return 0x80;
			}
			
			fwrite(OutBuf, 0x01, OutLen, hFile);
			fclose(hFile);
			
			if (CurFile == 0x0001143A && ! FileCount)
			{
				FileCount = ReadTOC(OutLen, OutBuf, &Files);
			}
		}
		
		free(OutBuf);
		
		if (CurFile == 0x00007F7F)
			break;
		
		CurPos += DataLen;
	} while(CurPos < FileLen);
	
	if (CurFile == 0x00007F7F)
	{
		printf("\n");
		printf("File %X Contents\n", CurFile);
		printf("------------------\n");
		// That is the archive with all files to install
		PakData = &FileData[CurPos + 0x05];
		
		/*
		// File 0001143A has the TOC

	Install Creator Install Data
Note: There are 2 files with the ID 0001143A. #1 and #2 specify the first and second one.
      The second TOC file is invalid.

Version 3b: (verified with v3.0.50 Pro)
		13000000 - File Count
Cmp		Len  ??       FilePos  CmprSize unused   UncmprSz ChkSum?? Flags    Flags 2? unused   unused   FileNum  ??       ??       ??       ??       ??       ??       3xString (1FileName  2Folder  3Link), 1-3 Program Folder/Start Menu (Folder/Link), 4-6 Desktop, 7-9 Quick Launch?
Zl	#1	6F00 02000000 00000000 68010000 00000000 70020000 00000000 00E04F00 00000000 00000000 00000000 00000000 -------- -------- -------- -------- -------- -------- 1"Uninstal.exe" 2"Install Creator" 3"Uninstall Install Creator.lnk" 4"" 5"" 6"" 7"" 8"" 9"" 
Zl	#2	6F00 02000000 68010000 68010000 00000000 70020000 00000000 00E04F00 00000000 00000000 00000000 00000000 -------- -------- -------- -------- -------- -------- 1"Uninstal.exe" 2"Install Creator" 3"Uninstall Install Creator.lnk" 4"" 5"" 6"" 7"" 8"" 9"" 
Bz	#1	8A00 00000000 D0020000 CCE30B00 00000000 48C41900 77FEFBFB 08E0CF04 00000200 00000000 00000000 01000000 79B30695 6DD6CB01 79B30695 6DD6CB01 D918A996 6DD6CB01 1"ic.exe" 2"Install Creator" 3"Install Creator.lnk" 4"" 5"" 6"Install Creator.lnk" 7"" 8"" 9""
xx	#2	8A00 01000000 9CE60B00 A0D50300 00000000 48C41900 52071D4F 08E0CF04 00000200 00000000 00000000 01000000 79651EA8 6DD6CB01 79651EA8 6DD6CB01 79419BA9 6DD6CB01 1"ic.exe" 2"Install Creator" 3"Install Creator.lnk" 4"" 5"" 6"Install Creator.lnk" 7"" 8"" 9""
Zl		5500 00000000 3CBC0F00 90F41200 00000000 772A1300 6FDAA7C8 08E00F00 00000000 00000000 00000000 02000000 A6E74A04 27F6C901 A6E74A04 27F6C901 BED8C8E8 B822C801 1"ic.chm" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
Zl		7400 00000000 CCB02200 23060000 00000000 08170000 D6F37A58 08E04F00 00000000 00000000 00000000 03000000 C60B5204 27F6C901 C60B5204 27F6C901 7F628B50 9A6BCB01 1"License.txt" 2"Install Creator" 3"License.lnk" 4"" 5"" 6"" 7"" 8"" 9"" 
Bz		7200 00000000 EFB62200 CF0F0000 00000000 56330000 32E69EDA 08E04F08 00000000 00000000 00000000 04000000 6C527727 E677CA01 6C527727 E677CA01 190AD07B 6CD6CB01 1"Readme.txt" 2"Install Creator" 3"Readme.lnk" 4"" 5"" 6"" 7"" 8"" 9"" 
Zl		6E00 00000000 BEC62200 5D020000 00000000 EB0A0000 161556A0 08E00F00 00000000 00000000 00000000 05000000 266D5404 27F6C901 266D5404 27F6C901 007843BA FF51C301 1"Templ\Classic - With Bitmap.imt" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
Zl		7100 00000000 1BC92200 0F020000 00000000 740A0000 B313824A 08E00F00 00000000 00000000 00000000 06000000 266D5404 27F6C901 266D5404 27F6C901 000DFEC4 FF51C301 1"Templ\Classic - Without Bitmap.imt" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
Bz		6400 00000000 2ACB2200 BC190000 00000000 365E0000 A6DC2956 08E00F00 00000000 00000000 00000000 07000000 266D5404 27F6C901 266D5404 27F6C901 00FB972D AA11BE01 1"Templ\ClassicLeft.bmp" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
Zl		6000 00000000 E6E42200 86020000 00000000 DE0A0000 65DF2DC1 08E00F00 00000000 00000000 00000000 08000000 266D5404 27F6C901 266D5404 27F6C901 00240B1F E3A9C301 1"Templ\Default.imt" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
Bz		5E00 00000000 6CE72200 C2060000 00000000 5F140000 4ACDF1EA 08E00F00 00000000 00000000 00000000 09000000 266D5404 27F6C901 266D5404 27F6C901 00F12622 90A9C601 1"Templ\Dutch.txt" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
Bz		5F00 00000000 2EEE2200 CF060000 00000000 7F140000 FA02397E 08E00F00 00000000 00000000 00000000 0A000000 86CE5604 27F6C901 86CE5604 27F6C901 00BE50B4 ECBCC301 1"Templ\French.txt" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
Bz		5F00 00000000 FDF42200 35070000 00000000 C8160000 F063E8C7 08E00F00 00000000 00000000 00000000 0B000000 86CE5604 27F6C901 86CE5604 27F6C901 BAC28003 0E2EC801 1"Templ\German.txt" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
Bz		6000 00000000 32FC2200 CE060000 00000000 83150000 B3980EBE 08E00F00 00000000 00000000 00000000 0C000000 86CE5604 27F6C901 86CE5604 27F6C901 00465AA1 8815C401 1"Templ\Italian.txt" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
Bz		6300 00000000 00032300 72420000 00000000 06B10000 546C667B 08E00F00 00000000 00000000 00000000 0D000000 E62F5904 27F6C901 E62F5904 27F6C901 00A6823B 2138C301 1"Templ\ModernLeft.bmp" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
Zl		6200 00000000 72452300 C1080000 00000000 92110000 EBB838EB 08E00F00 00000000 00000000 00000000 0E000000 E62F5904 27F6C901 E62F5904 27F6C901 0003A129 2138C301 1"Templ\ModernTop.bmp" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
Bz		6000 00000000 334E2300 C5060000 00000000 98130000 74E42BB0 08E00F00 00000000 00000000 00000000 0F000000 E62F5904 27F6C901 E62F5904 27F6C901 004D0C4C 67BEC401 1"Templ\Spanish.txt" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
Bz		6000 00000000 F8542300 D6060000 00000000 C2130000 72C9913A 08E00F00 00000000 00000000 00000000 10000000 E62F5904 27F6C901 E62F5904 27F6C901 0050ABC6 0322C601 1"Templ\Swedish.txt" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
Zl		5D00 00000000 CE5B2300 9E000000 00000000 66010000 30F4B753 08E00F00 00000000 00000000 00000000 11000000 8697F703 27F6C901 8697F703 27F6C901 00187157 9335C301 1"UserImages.bmp" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
Zl	#1	5B00 00000000 6C5C2300 F4170000 00000000 E8170000 00B514E9 08E00F00 00000000 00000000 00000000 12000000 E6F8F903 27F6C901 E6F8F903 27F6C901 56C6D751 A8E8C701 1"settings.dat" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
xx	#2	5B00 01000000 60742300 1D000000 00000000 E8170000 49E31481 08E00F00 00000000 00000000 00000000 12000000 E6F8F903 27F6C901 E6F8F903 27F6C901 107E0012 FA6EC701 1"settings.dat" 2"" 3"" 4"" 5"" 6"" 7"" 8"" 9"" 
		
		Flags: Bit 3 (08):	has additional (File Date?) fields (0 - no, 1 - yes)

Version 2b: (verified with v2.0.35 Free)
		B3050000 - File Count
		Len      ??       Flags    unused   unused   ChkSum?  UncmprSz FilePos  CmprSize unused
		85000000 02000000 00E00F00 00000000 00000000 00000000 5F020000 00000000 4E010000 00000000 "Uninstal.exe" "Sonic Arena The Lost Chapters" "Uninstall Sonic Arena The Lost Chapters.lnk"
		55000000 00000000 08E40F00 00000000 00000000 A03E9444 622C0000 4E010000 8D0B0000 00000000 E05ECAD9 5211CB01 0098CDBD 0447CB01 00FC3233 5A73C901 "mingwm10.dll"
		5B000000 00000000 08E40F00 00000000 00000000 10251DE2 00A80000 DB0C0000 9D4F0000 00000000 40D8C8D9 5211CB01 0098CDBD 0447CB01 00657DA9 B4F3C901 "libgcc_s_dw2-1.dll"
		54000000 00000000 08E40F00 00000000 00000000 FF4FF030 00722400 785C0000 FE8D0E00 00000000 80E5CBD9 5211CB01 0098CDBD 0447CB01 001E37E7 B780CA01 "QtCore4.dll"
		5D000000 00000000 08E00F00 00000000 00000000 BB4BB1A0 7ED70000 76EA0E00 EE130000 00000000 80D3441F AFEFCA01 0098CDBD 0447CB01 00D8A63C 4931C801 "Audio\BGM\Field2.mid"
		5D000000 00000000 08E00F00 00000000 00000000 BFD2A29B 93060100 64FE0E00 25110000 00000000 A0744C1F AFEFCA01 0098CDBD 0447CB01 00D8A63C 4931C801 "Audio\BGM\Field3.mid"

Version 2a: (verified with v2.0.16 Free)
		35000000 - File Count
		Len      ??       Flags    unused   ChkSum?  UncmprSz FilePos  CmprSize
		63000000 02000000 00E00700 00000000 00000000 3E020000 00000000 45010000 "Uninstal.exe" "Neo Sonic Universe" "Uninstall Neo Sonic Universe"
		4C000000 00000000 08E40700 00000000 4F17B1D9 005C0400 45010000 D2160200 204BEDF2 2506C301 00B8B818 7F8CC301 00CCBC31 BF77C101 "Cncs232.dll"
		8D000000 00000000 08E00704 00000000 EE3FE096 16550C00 17180200 8AEC0400 60DBC02E 6F1EC301 00B8B818 7F8CC301 00EC61FB A18BC301 "Neo Sonic Universe.exe" "Neo Sonic Universe" "Neo Sonic Universe" "" "" "Neo Sonic Universe" "" "" ""


Version 3a: (verified with v2.0.24 Pro and v2.0.31 Pro)
		13000000 - File Count
Cmp		Len  ??       FilePos  CmprSize unused   UncmprSz ChkSum?? Flags    Flags 2? unused   unused   FileNum
		7500 02000000 6A010000 6A010000 00000000 7C020000 00000000 00E04700 00000000 00000000 0000     00000000 "Uninstal.exe" "Install Creator Pro" "Uninstall Install Creator Pro.lnk"
		9500 01000000 408A0D00 A34B0500 00000000 00C01C00 602B963C 09E0C704 00000000 00000000 0000     01000000 F0721AB9 16E4C401 71FA8874 0723C501 600986BC 16E4C401 "icp.exe" "Install Creator Pro" "Install Creator Pro.lnk" "" "" "Install Creator Pro.lnk"
		5400 00000000 E3D51200 A9611800 00000000 95951800 834BFF34 08E00700 00000000 00000000 0000     02000000 9083C8BE 9D4BC301 51CB6876 0723C501 20607452 15CFC401 "icp.chm"
		*/
		for (CurFile = 0x00; CurFile < FileCount; CurFile ++)
		{
			TempFile = &Files[CurFile];
			printf("File %u\n", TempFile->FileID);
			/*printf("    Location: Main (Program Folder / Start Menu):\n");
			printf("        Name:   %s\n", TempFile->Loc_Main.Name);
			printf("        Link:   %s\\%s\n", TempFile->Loc_Main.Folder, TempFile->Loc_Main.Link);
			printf("    Location Desktop:\n");
			printf("        Name:   %s\n", TempFile->Loc_Desktop.Name);
			printf("        Link:   %s\\%s\n", TempFile->Loc_Desktop.Folder, TempFile->Loc_Desktop.Link);
			printf("    Location Quick Launch:\n");
			printf("        Name:   %s\n", TempFile->Loc_QLaunch.Name);
			printf("        Link:   %s\\%s\n", TempFile->Loc_QLaunch.Folder, TempFile->Loc_QLaunch.Link);*/
			printf("    Name: %s\n", TempFile->Loc_Main.Name);
			
			OutLen = TempFile->FileSize;
			RetVal = UnpackFile(&PakData[TempFile->DataPos], TempFile->DataSize, &OutBuf, &OutLen);
			
			if (! RetVal)
			{
				sprintf(OutFileName, "%s\\Program\\%s", Path, TempFile->Loc_Main.Name);
				switch(FileData[CurPos + 0x04])
				{
				case 0x01:
				case 0x02:
					break;
				default:
					strcat(OutFileName, ".cmp");
					break;
				}
				CreatePath(OutFileName);
				hFile = fopen(OutFileName, "wb");
				if (hFile == NULL)
				{
					printf("Error: Can't open Output File!\n");
					return 0x80;
				}
				
				fwrite(OutBuf, 0x01, OutLen, hFile);
				fclose(hFile);
			}
			
			free(OutBuf);
		}
	}
	
	free(OutFileName);
	free(Files);
	
	printf("Done.\n");
	
	return 0x00;
}

UINT32 ReadTOC(const UINT32 TOCLen, const char* TOCData, INS_CRT_FILE** RetFileStruct)
{
	UINT32 FileCount;
	INS_CRT_FILE* Files;
	UINT32 BasePos;
	UINT32 CurPos;
	UINT32 CurFile;
	INS_CRT_FILE* TempFile;
	INS_CRT_FILE_TIMES* TempTime;
	UINT8 FileVer;
	
	BasePos = 0x00;
	memcpy(&FileCount, &TOCData[BasePos], 0x04);
	Files = (INS_CRT_FILE*)malloc(FileCount * sizeof(INS_CRT_FILE));
	*RetFileStruct = Files;
	BasePos += 0x04;
	
	if (! TOCData[BasePos + 0x02])
	{
		memcpy(&CurPos, &TOCData[BasePos + 0x14], 0x04);
		if (CurPos)
			FileVer = 0x01;
		else
			FileVer = 0x02;
	}
	else
	{
		if (TOCData[BasePos + 0x2C])
			FileVer = 0x11;
		else
			FileVer = 0x12;
	}
	printf("\tTOC Version: %s v%u\n", (FileVer & 0x10) ? "Pro" : "Free", FileVer & 0x0F);
	
	for (CurFile = 0x00; CurFile < FileCount; CurFile ++)
	{
		if (BasePos >= TOCLen)
			return CurFile;
		
		TempFile = &Files[CurFile];
		CurPos = BasePos;
		switch(FileVer)
		{
		case 0x01:	// Free v1
			memcpy(&TempFile->EntryLen,		&TOCData[CurPos + 0x00], 0x04);
			memcpy(&TempFile->Unknown02,	&TOCData[CurPos + 0x04], 0x04);
			memcpy(&TempFile->Flags1,		&TOCData[CurPos + 0x08], 0x04);
			memcpy(&TempFile->Flags2,		&TOCData[CurPos + 0x0C], 0x04);
			memcpy(&TempFile->ChkSum,		&TOCData[CurPos + 0x10], 0x04);
			memcpy(&TempFile->FileSize,		&TOCData[CurPos + 0x14], 0x04);
			memcpy(&TempFile->DataPos,		&TOCData[CurPos + 0x18], 0x04);
			memcpy(&TempFile->DataSize,		&TOCData[CurPos + 0x1C], 0x04);
			
					TempFile->Unused0E =	0x00;
					TempFile->Unused22 =	0x00;
					TempFile->Unused26 =	0x00;
					TempFile->FileID =		CurFile;
			CurPos += 0x20;
			break;
		case 0x02:	// Free v2
			memcpy(&TempFile->EntryLen,		&TOCData[CurPos + 0x00], 0x04);
			memcpy(&TempFile->Unknown02,	&TOCData[CurPos + 0x04], 0x04);
			memcpy(&TempFile->Flags1,		&TOCData[CurPos + 0x08], 0x04);
			memcpy(&TempFile->Flags2,		&TOCData[CurPos + 0x0C], 0x04);
			memcpy(&TempFile->Unused22,		&TOCData[CurPos + 0x10], 0x04);
			memcpy(&TempFile->ChkSum,		&TOCData[CurPos + 0x14], 0x04);
			memcpy(&TempFile->FileSize,		&TOCData[CurPos + 0x18], 0x04);
			memcpy(&TempFile->DataPos,		&TOCData[CurPos + 0x1C], 0x04);
			memcpy(&TempFile->DataSize,		&TOCData[CurPos + 0x20], 0x04);
			memcpy(&TempFile->Unused26,		&TOCData[CurPos + 0x24], 0x04);
			
					TempFile->Unused0E =	0x00;
					TempFile->FileID =		CurFile;
			CurPos += 0x28;
			break;
		case 0x11:	// Pro v1
			memcpy(&TempFile->EntryLen,		&TOCData[CurPos + 0x00], 0x02);
			TempFile->EntryLen &= 0x0000FFFF;
			memcpy(&TempFile->Unknown02,	&TOCData[CurPos + 0x02], 0x04);
			memcpy(&TempFile->DataPos,		&TOCData[CurPos + 0x06], 0x04);
			memcpy(&TempFile->DataSize,		&TOCData[CurPos + 0x0A], 0x04);
			memcpy(&TempFile->Unused0E,		&TOCData[CurPos + 0x0E], 0x04);
			memcpy(&TempFile->FileSize,		&TOCData[CurPos + 0x12], 0x04);
			memcpy(&TempFile->ChkSum,		&TOCData[CurPos + 0x16], 0x04);
			memcpy(&TempFile->Flags1,		&TOCData[CurPos + 0x1A], 0x04);
			memcpy(&TempFile->Flags2,		&TOCData[CurPos + 0x1E], 0x04);
			memcpy(&TempFile->Unused22,		&TOCData[CurPos + 0x22], 0x04);
			memcpy(&TempFile->Unused26,		&TOCData[CurPos + 0x26], 0x02);
					TempFile->Unused26 &= 0x0000FFFF;
			memcpy(&TempFile->FileID,		&TOCData[CurPos + 0x28], 0x04);
			CurPos += 0x2C;
			break;
		case 0x12:	// Pro v2
			memcpy(&TempFile->EntryLen,		&TOCData[CurPos + 0x00], 0x02);
			TempFile->EntryLen &= 0x0000FFFF;
			memcpy(&TempFile->Unknown02,	&TOCData[CurPos + 0x02], 0x04);
			memcpy(&TempFile->DataPos,		&TOCData[CurPos + 0x06], 0x04);
			memcpy(&TempFile->DataSize,		&TOCData[CurPos + 0x0A], 0x04);
			memcpy(&TempFile->Unused0E,		&TOCData[CurPos + 0x0E], 0x04);
			memcpy(&TempFile->FileSize,		&TOCData[CurPos + 0x12], 0x04);
			memcpy(&TempFile->ChkSum,		&TOCData[CurPos + 0x16], 0x04);
			memcpy(&TempFile->Flags1,		&TOCData[CurPos + 0x1A], 0x04);
			memcpy(&TempFile->Flags2,		&TOCData[CurPos + 0x1E], 0x04);
			memcpy(&TempFile->Unused22,		&TOCData[CurPos + 0x22], 0x04);
			memcpy(&TempFile->Unused26,		&TOCData[CurPos + 0x26], 0x04);
			memcpy(&TempFile->FileID,		&TOCData[CurPos + 0x2A], 0x04);
			CurPos += 0x2E;
			break;
		}
		
		TempTime = &TempFile->FileTimes;
		if (TempFile->Flags1 & 0x00000008)
		{
			memcpy(&TempTime->Time1,	&TOCData[CurPos + 0x00], 0x08);
			memcpy(&TempTime->Time2,	&TOCData[CurPos + 0x08], 0x08);
			memcpy(&TempTime->Time3,	&TOCData[CurPos + 0x10], 0x08);
			CurPos += 0x18;
		}
		else
		{
			memset(&TempTime->Time1, 0x00, 0x08);
			memset(&TempTime->Time2, 0x00, 0x08);
			memset(&TempTime->Time3, 0x00, 0x08);
		}
		
		CurPos += ReadString(&TempFile->Loc_Main.Name,		&TOCData[CurPos]);
		CurPos += ReadString(&TempFile->Loc_Main.Folder,	&TOCData[CurPos]);
		CurPos += ReadString(&TempFile->Loc_Main.Link,		&TOCData[CurPos]);
		CurPos += ReadString(&TempFile->Loc_Desktop.Name,	&TOCData[CurPos]);
		CurPos += ReadString(&TempFile->Loc_Desktop.Folder,	&TOCData[CurPos]);
		CurPos += ReadString(&TempFile->Loc_Desktop.Link,	&TOCData[CurPos]);
		CurPos += ReadString(&TempFile->Loc_QLaunch.Name,	&TOCData[CurPos]);
		CurPos += ReadString(&TempFile->Loc_QLaunch.Folder,	&TOCData[CurPos]);
		CurPos += ReadString(&TempFile->Loc_QLaunch.Link,	&TOCData[CurPos]);
		
		BasePos += TempFile->EntryLen;
	}
	
	return FileCount;
}

INLINE UINT32 ReadString(char** String, const char* FileData)
{
	UINT32 StrLen;
	
	StrLen = strlen(FileData) + 0x01;
	*String = malloc(StrLen);
	strcpy(*String, FileData);
	
	return StrLen;
}

void CreatePath(const char* FileName)
{
	char* Path;
	char ChrBak;
	char* TempStr;
	
	Path = (char*)malloc(strlen(FileName) + 0x01);
	strcpy(Path, FileName);
	TempStr = strchr(Path, '\\');
	while(TempStr != NULL)
	{
		ChrBak = *TempStr;
		*TempStr = 0x00;
		CreateDirectory(Path, NULL);
		*TempStr = ChrBak;
		TempStr ++;
		
		TempStr = strchr(TempStr, '\\');
	}
	
	return;
}

UINT8 UnpackFile(const UINT8* Data, UINT32 DataLen, char** OutBuf, UINT32* OutLen)
{
	int RetVal;
	UINT8 ComprType;
	char* ComprStr;
	
	ComprType = *(Data - 0x01);	// the Install Creator really seems to handle it like this
	switch(ComprType)
	{
	case 0x01:
		ComprStr = "zlib";
		printf("    Compression: %02X (%s)\n", ComprType, ComprStr);
		printf("    Size Compressed: %u, Uncompressed: %u, Ratio: %.2f%%\n", DataLen, *OutLen,
				100.0 * DataLen / *OutLen);
		
		*OutLen = *OutLen;
		*OutBuf = (UINT8*)malloc(*OutLen);
		
		printf("    Decompressing ...");
		RetVal = uncompress(*OutBuf, OutLen, Data, DataLen);
		printf("  ");
		switch(RetVal)
		{
		case Z_OK:
			printf("OK");
			break;
		case Z_MEM_ERROR:
			printf("Memory Error");
			break;
		case Z_BUF_ERROR:
			printf("Output Buffer too small");
			break;
		case Z_DATA_ERROR:
			printf("Input Data Error");
			break;
		}
		printf("!\n");
		break;
	case 0x02:
		ComprStr = "BZip2";
		printf("    Compression: %02X (%s)\n", ComprType, ComprStr);
		printf("    Size Compressed: %u, Uncompressed: %u, Ratio: %.2f%%\n", DataLen, *OutLen,
				100.0 * DataLen / *OutLen);
		
		*OutLen = *OutLen;
		*OutBuf = (UINT8*)malloc(*OutLen);
		
		printf("    Decompressing ...");
		RetVal = BZ2_bzBuffToBuffDecompress(*OutBuf, OutLen, Data, DataLen, 0, 0);
		printf("  ");
		switch(RetVal)
		{
		case BZ_OK:
			printf("OK");
			break;
		case BZ_MEM_ERROR:
			printf("Memory Error");
			break;
		case BZ_OUTBUFF_FULL:
			printf("Output Buffer too small");
			break;
		case BZ_DATA_ERROR:
			printf("Input Data Error");
			break;
		}
		printf("!\n");
		break;
	default:
		ComprStr = "Unknown";
		printf("    Compression: %02X (%s)\n", ComprType, ComprStr);
		printf("    Size Compressed: %u, Uncompressed: %u, Ratio: %.2f%%\n", DataLen, *OutLen,
				100.0 * DataLen / *OutLen);
		
		*OutLen = DataLen;
		*OutBuf = (UINT8*)malloc(*OutLen);
		memcpy(*OutBuf, Data, DataLen);
		break;
	}
	
	return RetVal & 0xFF;
}
