// Stand Alone Game Installer Decompressor
//
// written by Valley Bell, August 2011
//
// 
// Note:
//  I reverse-engineered almost all functions (except for File I/O and MoveElements) by
//  disassembling an Installer EXE.
//  Addresses in comments refer to the RAM offset of the function in the Install EXE.
//  I used an EXE written with The Games Factory 1.7.1.38 (32-bit).

#include <malloc.h>

// Type Definitions for short types
typedef unsigned char	UINT8;
typedef signed char 	INT8;

typedef unsigned short	UINT16;
typedef signed short	INT16;

typedef unsigned int	UINT32;
typedef signed int		INT32;

typedef unsigned char	bool;
#define false			0
#define true			1

#define NULL	((void*)0)


// Function Prototypes
void InitTables(void);			 		 		 		 				// 00402519

static UINT16 ReadByte(void);		 		 		 		 			// 00401D35
static void RecalculateTables(void); 		 		 		 			// 0040216C
static void ProcessByteA(UINT16 Arg); 		 		 		 			// 004022BE
static UINT16 ReadValueBit(void); 	 		 		 		 			// 0040239D
static UINT16 ReadValueByte(void); 		 		 		 				// 00402409
static UINT16 ProcessByteB(void);	 		 		 		 			// 004024B0
static void MoveElements(UINT16* DstArr, UINT16* SrcArr, UINT32 Count);	// 00404A80


// Decompression Tables

// static tables (stored in EXE at offset 0x6CE0)
static UINT8 Step1[0x100];
static UINT8 Step2[0x100];
static bool InitStaticTbls = false;

// dynamic talbles
static UINT16 Mask[0x274];
static UINT16 Tbl1[0x273];
static UINT16 Tbl2[0x3AD];
static UINT16* Tbl3 = &Tbl2[0x273];

static UINT8 Buffer[0x1000];

// global variables
UINT32 SrcPos;
UINT32 SrcLen;
UINT8* SrcData;
static UINT16 Var44;	// Variable at 0x40AF44
static UINT16 VarD4;	// Variable at 0x4090D4
static UINT16 VarD8;	// Variable at 0x4090D8

void InitStaticTables(void)
{
	UINT16 CurIdx;
	UINT8 CurVal;
	
	// actually this table is stored in pregenerated form in the EXE
	CurVal = 0x00;
	for (CurIdx = 0x00; CurIdx < 0x20; CurIdx ++)
	{
		Step1[CurIdx] = CurVal;
		Step2[CurIdx] = 0x03;
	}
	for (; CurIdx < 0x50; CurIdx ++)
	{
		if (! (CurIdx & 0x0F))
			CurVal ++;
		Step1[CurIdx] = CurVal;
		Step2[CurIdx] = 0x04;
	}
	for (; CurIdx < 0x90; CurIdx ++)
	{
		if (! (CurIdx & 0x07))
			CurVal ++;
		Step1[CurIdx] = CurVal;
		Step2[CurIdx] = 0x05;
	}
	for (; CurIdx < 0xC0; CurIdx ++)
	{
		if (! (CurIdx & 0x03))
			CurVal ++;
		Step1[CurIdx] = CurVal;
		Step2[CurIdx] = 0x06;
	}
	for (; CurIdx < 0xF0; CurIdx ++)
	{
		if (! (CurIdx & 0x01))
			CurVal ++;
		Step1[CurIdx] = CurVal;
		Step2[CurIdx] = 0x07;
	}
	for (; CurIdx < 0x100; CurIdx ++)
	{
		CurVal ++;
		Step1[CurIdx] = CurVal;
		Step2[CurIdx] = 0x08;
	}
	
	InitStaticTbls = true;
	
	return;
}

UINT8 DecompressData(UINT32* DecLen, UINT8** DecData, UINT32 CmpLen, UINT8* CmpData)
{
	UINT32 DstPos;
	UINT32 DstLen;
	UINT8* DstData;
	UINT16 BufPos;
	UINT16 TmpVar1;
	UINT16 TmpVar2;
	UINT16 TmpVar3;
	
	UINT16 AX;
	UINT16 CX;
	UINT16 SI;
	UINT16 BX;
	
	if (! InitStaticTbls)
		InitStaticTables();
	
	SrcLen = CmpLen;
	SrcData = CmpData;
	
	DstLen = *(UINT32*)(SrcData + 0x00);
	DstData = *DecData;
	if (DstData == NULL || *DecLen != DstLen)
	{
		if (DstData != NULL)
			free(DstData);
		*DecLen = DstLen;
		
		DstData = (UINT8*)malloc(DstLen);
		*DecData = DstData;
		if (DstData == NULL)
			return 0x80;	// Allocation Error
	}
	
	
	InitTables();
	
	
	Var44 = 0x270;
	VarD4 = 0x000;
	VarD8 = 0x000;
	BufPos = 0xFC4;
	
	SrcPos = 0x04;
	DstPos = 0x00;
	while(DstPos < DstLen)
	{
		SI = Var44;
		
		while(SI < 0x273)
		{
			AX = ReadValueBit();
			SI = Tbl1[SI + AX];
		}
		SI -= 0x273;
		
		ProcessByteA(SI);
		
		if (SI < 0x100)
		{
			DstData[DstPos] = (UINT8)SI;
			DstPos ++;
			Buffer[BufPos] = (UINT8)SI;
			BufPos ++;
			BufPos &= 0xFFF;
		}
		else
		{
			AX = ProcessByteB();
			
			// 00402619
			TmpVar1 = 0x00;
			TmpVar3 = (BufPos - AX - 0x01) & 0xFFF;
			BX = SI - 0xFD;
			while(TmpVar1 < BX)
			{
				// 0040263A
				CX = (TmpVar1 + TmpVar3) & 0xFFF;
				TmpVar2 = Buffer[CX];
				
				DstData[DstPos] = (UINT8)TmpVar2;
				DstPos ++;
				
				Buffer[BufPos] = (UINT8)TmpVar2;
				BufPos ++;
				BufPos &= 0xFFF;
				TmpVar1 ++;
			}
		}
	}
	
	return 0x00;
}

void InitTables(void)	// 00402519
{
	UINT16 CurIdx;
	UINT16 IdxVal;
	
	for (CurIdx = 0x00; CurIdx < 0x13A; CurIdx ++)
	{
		Mask[CurIdx] = 0x01;
		Tbl1[CurIdx] = 0x273 + CurIdx;
		Tbl3[CurIdx] = CurIdx;
	}
	IdxVal = 0x00;
	for (CurIdx = 0x13A; CurIdx < 0x273; CurIdx ++)
	{
		Mask[CurIdx] = Mask[IdxVal + 0x01] + Mask[IdxVal + 0x00];
		Tbl1[CurIdx] = IdxVal;
		Tbl2[IdxVal + 0x00] = CurIdx;
		Tbl2[IdxVal + 0x01] = CurIdx;
		IdxVal += 0x02;
	}
	Mask[0x273] = 0xFFFF;
	Tbl2[0x272] = 0x000;
	
	
	for (CurIdx = 0x00; CurIdx < 0xFC4; CurIdx ++)
		Buffer[CurIdx] = 0x20;
	for (; CurIdx < 0x1000; CurIdx ++)
		Buffer[CurIdx] = 0x00;
	
	return;
}

static UINT16 ReadByte(void)	// 00401D35
{
	// return values should be the same as in the original decompression
	if (SrcData == NULL)
		return 0xFFFE;
	if (SrcPos >= SrcLen)
		return 0xFFFF;
	
	return SrcData[SrcPos ++];
}

static void RecalculateTables(void)	// 0040216C
{
	UINT16 CX;
	UINT16 SI;
	UINT16 DI;
	UINT16 BX;
	INT16 AX;
	
	// 0040216C
	SI = 0x00;
	for (CX = 0x000; CX < 0x273; CX ++)
	{
		BX = Tbl1[CX];
		if (BX >= 0x273)
		{
			AX = Mask[CX] + 0x01;
			Tbl1[SI] = BX;
			if (AX & 0x8000)
			{
				//Stop
				AX ++;
			}
			AX >>= 1;	// signed bit-shift
			Mask[SI] = AX;
			SI ++;
		}
	}
	
	// 004021C3
	SI = 0x00;
	for (DI = 0x13A; DI < 0x273; DI ++)
	{
		BX = Mask[SI + 0x01] + Mask[SI + 0x00];
		Mask[DI] = BX;
		CX = DI - 0x01;
		while(Mask[CX] > BX)
			CX --;
		CX ++;
		
		AX = DI - CX;
		
		MoveElements(Mask + (CX + 0x01), Mask + CX, AX);
		Mask[CX] = BX;
		MoveElements(Tbl1 + (CX + 0x01), Tbl1 + CX, AX);
		Tbl1[CX] = SI;
		
		SI += 0x02;
	}
	
	// 0040227A
	for (CX = 0x00; CX < 0x273; CX ++)
	{
		AX = Tbl1[CX];
		if (AX < 0x273)
		{
			Tbl2[AX + 0x00] = CX;
			Tbl2[AX + 0x01] = CX;
		}
		else
		{
			Tbl2[AX] = CX;
		}
	}
	
	return;
}

static void ProcessByteA(UINT16 Arg)	// 004022BE
{
	UINT16 CX;
	UINT16 SI;
	UINT16 DX;
	UINT16 DI;
	UINT16 BX;
	
	if (Mask[0x272] == 0x8000)
	{
		RecalculateTables();
	}
	
	CX = Tbl3[Arg];
	do
	{
		DI = CX;
		DX = CX + 1;
		
		SI = Mask[DI] + 0x01;
		Mask[DI] = SI;
		if (Mask[DX] < SI)
		{
			do
			{
				DX ++;
			} while(Mask[DX] < SI);
			DX --;
			
			Mask[DI] = Mask[DX];
			Mask[DX] = SI;
			
			SI = DI;
			BX = Tbl1[SI];
			Tbl2[BX] = DX;	// Note: actually accesses Tbl3
			if (BX < 0x273)
				Tbl2[BX + 1] = DX;
			DI = Tbl1[DX];
			Tbl1[DX] = BX;
			
			Tbl2[DI] = CX;
			if (DI < 0x273)
				Tbl2[DI + 1] = CX;
			CX = DX;
			Tbl1[SI] = DI;
		}
		
		CX = Tbl2[CX];
	} while(CX);
	
	return;
}

static UINT16 ReadValueBit(void)	// 0040239D
{
	UINT16 AX;
	UINT8 CL;
	
	while(VarD8 <= 0x08)
	{
		AX = ReadByte();
		if (AX == 0xFFFE)
			return 0xFFFE;
		else if (AX == 0xFFFF)
			AX = 0x00;
		CL = 0x08 - VarD8;
		
		AX <<= CL;
		VarD8 += 0x08;
		VarD4 |= AX;
	}
	VarD8 --;
	
	AX = VarD4 >> 15;
	VarD4 *= 2;
	
	return AX;
}

static UINT16 ReadValueByte(void) // 00402409
{
	UINT16 AX;
	UINT8 CL;
	
	while(VarD8 <= 0x08)
	{
		AX = ReadByte();
		if (AX == 0xFFFE)
			return 0xFFFE;
		else if (AX == 0xFFFF)
			AX = 0x00;
		CL = 0x08 - VarD8;
		
		AX <<= CL;
		VarD8 += 0x08;
		VarD4 |= AX;
	}
	VarD8 -= 8;
	
	AX = VarD4 >> 8;
	VarD4 <<= 8;
	
	return AX;
}

static UINT16 ProcessByteB(void)	// 004024B0
{
	UINT16 AX;
	UINT16 SI;
	UINT16 DI;
	UINT16 BX;
	
	AX = ReadValueByte();
	if (AX == 0xFFFE)
		return 0xFFFE;
	
	SI = AX;
	DI = Step1[AX] << 6;
	AX = Step2[AX] - 0x02;
	BX = AX - 0x01;
	while(AX)
	{
		AX = ReadValueBit();
		SI = AX + SI * 2;
		AX = BX;
		BX --;
	}
	
	return (SI & 0x3F) | DI;
}

static void MoveElements(UINT16* DstArr, UINT16* SrcArr, UINT32 Count)	// 00404A80
{
	UINT32 CurEl;
	UINT16* DstPnt;
	UINT16* SrcPnt;
	
	SrcPnt = SrcArr + Count;
	DstPnt = DstArr + Count;
	for (CurEl = 0x00; CurEl < Count; CurEl ++)
	{
		SrcPnt --;
		DstPnt --;
		*DstPnt = *SrcPnt;
	}
		
	return;
}
