// Unpack Stand Alone Game Installer

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX_PATH          260	// from windef.h


// Type Definitions for short types
typedef unsigned char	UINT8;
typedef signed char 	INT8;

typedef unsigned short	UINT16;
typedef signed short	INT16;

typedef unsigned int	UINT32;
typedef signed int		INT32;


#ifdef WIN32
#define DIR_CHR		'\\'
#define DIR_STR		"\\"
#else
#define DIR_CHR		'/'
#define DIR_STR		"/"
#endif


#define FCC_INSA	0x41534E49


// Structures
typedef struct sagi_file_list
{
	UINT16 FolderID;
	UINT16 ID;
	char* FileName;
	struct
	{
		UINT16 Unknown1;
		UINT16 Unknown2;
		UINT16 Unknown3;
		UINT16 Unknown4;
	} FolderAttr;
	UINT16 Unknown;
	UINT32 DataPos;
	UINT32 DataSize;
} SAGI_FILE;


// Function Prototypes
int main(int argc, char* argv[]);
UINT8 UnpackSAGInstaller(char* File, char* Path);
void UnpackFile(UINT8* Data, UINT32 DataLen, UINT8** OutBuf, UINT32* OutLen);

// Prototype 
UINT8 DecompressData(UINT32* DecLen, UINT8** DecData, UINT32 CmpLen, UINT8* CmpData);

int main(int argc, char* argv[])
{
	char* InstallFile;
	char* ExtractPath;
	UINT8 RetVal;
	
	printf("SAG Installer Data Extractor\n----------------------------\n");
	if (argc <= 0x02)
	{
		printf("Usage: SAGIExtract.exe setup.bin ExtractPath\n");
		return 1;	// need a file
	}
	
	InstallFile = argv[0x01];
	ExtractPath = (char*)malloc(strlen(argv[0x02]) + 0x02);
	strcpy(ExtractPath, argv[0x02]);
	if (ExtractPath[strlen(ExtractPath) - 0x01] != DIR_CHR)
		strcat(ExtractPath, DIR_STR);
	
	RetVal = UnpackSAGInstaller(InstallFile, ExtractPath);
	if (RetVal)
		return RetVal;
	
#ifdef _DEBUG
	getchar();
#endif
	
	return 0;
}

UINT8 UnpackSAGInstaller(char* File, char* Path)
{
	FILE* hFile;
	FILE* hFileDat;
	UINT32 FileLen;
	UINT8* FileData;
	UINT32 CurPos;
	UINT16 FileCount;
	SAGI_FILE* Files;
	UINT16 CurFile;
	char DatFile[MAX_PATH];
	char GameName[0x50];
	UINT16 EntryLen;
	UINT32 TempLng;
	char* TempStr;
	SAGI_FILE* TempFile;
	UINT32 OutLen;
	UINT8* OutBuf;
	char* OutFileName;
	
	hFile = fopen(File, "rb");
	if (hFile == NULL)
	{
		printf("Error: Can't open Installer File!\n");
		return 0x10;
	}
	
	fread(&TempLng, 0x04, 0x01, hFile);
	if (TempLng != FCC_INSA)
	{
		printf("Error: Invalid Installer File!\n");
		return 0x20;
	}
	
	fseek(hFile, 0x00, SEEK_END);
	FileLen = ftell(hFile);
	
	fseek(hFile, 0x00, SEEK_SET);
	FileData = (UINT8*)malloc(FileLen);
	fread(FileData, 0x01, FileLen, hFile);
	
	fclose(hFile);
	
	CurPos = 0x04;
	strncpy(DatFile, (char*)&FileData[CurPos], MAX_PATH);
	DatFile[MAX_PATH - 5] = 0x00;	// -5 to have space for ".bin"
	CurPos += MAX_PATH;
	
	// make dat-name (verified behaviour)
	TempStr = strrchr(DatFile, '.');
	if (TempStr == NULL)
		TempStr = DatFile + strlen(DatFile);
	strcpy(TempStr, ".dat");
	
	strncpy(GameName, (char*)&FileData[CurPos], 0x50);
	GameName[0x4F] = 0x00;
	CurPos += 0x50;
	
	printf("Game Name:  %s\n", GameName);
	printf("Data File:  %s\n", DatFile);
	
	TempStr = strrchr(File, DIR_CHR);
	if (TempStr == NULL)
		TempLng = 0;
	else
		TempLng = TempStr - File + 1;
	
	if (! TempLng)
	{
		hFileDat = fopen(DatFile, "rb");
	}
	else
	{
		TempStr = (char*)malloc(TempLng + strlen(DatFile) + 0x01);
		strncpy(TempStr, File, TempLng);
		strcpy(TempStr + TempLng, DatFile);
		hFileDat = fopen(TempStr, "rb");
		free(TempStr);
	}
	if (hFileDat == NULL)
	{
		printf("Error: Can't open Data File!\n");
		return 0x11;
	}
	printf("\n");
	
	// Read File TOC
	memcpy(&FileCount, &FileData[CurPos], 0x02);
	Files = (SAGI_FILE*)malloc(FileCount * sizeof(SAGI_FILE));
	CurPos += 0x02;
	printf("Files: %hu\n", FileCount);
	
	for (CurFile = 0x00; CurFile < FileCount; CurFile ++)
	{
		TempFile = &Files[CurFile];
		memcpy(&EntryLen,			&FileData[CurPos + 0x00], 0x02);
		memcpy(&TempFile->FolderID,	&FileData[CurPos + 0x02], 0x02);
		memcpy(&TempFile->ID,		&FileData[CurPos + 0x04], 0x02);
		TempLng = strlen((char*)&FileData[CurPos + 0x06]) + 0x02;
		TempFile->FileName = (char*)malloc(TempLng);
		memcpy(TempFile->FileName,	&FileData[CurPos + 0x06], TempLng);
		
		TempLng = 0x06 + TempLng;
		if (TempFile->FolderID)
		{
			memcpy(&TempFile->FolderAttr.Unknown1,	&FileData[CurPos + TempLng + 0x00], 0x02);
			memcpy(&TempFile->FolderAttr.Unknown2,	&FileData[CurPos + TempLng + 0x02], 0x02);
			memcpy(&TempFile->FolderAttr.Unknown3,	&FileData[CurPos + TempLng + 0x04], 0x02);
			memcpy(&TempFile->FolderAttr.Unknown4,	&FileData[CurPos + TempLng + 0x06], 0x02);
			TempLng += 0x08;
		}
		memcpy(&TempFile->Unknown,	&FileData[CurPos + TempLng + 0x00], 0x02);
		memcpy(&TempFile->DataPos,	&FileData[CurPos + TempLng + 0x02], 0x04);
		memcpy(&TempFile->DataSize,	&FileData[CurPos + TempLng + 0x06], 0x04);
		
		CurPos += EntryLen;
	}
	free(FileData);	FileData = NULL;
	
	OutFileName = (char*)malloc(strlen(Path) + MAX_PATH);
	for (CurFile = 0x00; CurFile < FileCount; CurFile ++)
	{
		TempFile = &Files[CurFile];
		
		switch(TempFile->FolderID)
		{
		case 0x0000:
			TempStr = "game file";
			break;
		case 0x0016:
			TempStr = "system file";
			break;
		default:
			TempStr = "unknown";
			break;
		}
		printf("%s\n", TempFile->FileName);
		printf("    File Type: 0x%04hX (%s)\n", TempFile->FolderID, TempStr);
		//printf("    Data Pos:\t0x%06u\n", TempFile->DataPos);
		
		FileLen = TempFile->DataSize;
		FileData = (UINT8*)malloc(FileLen);
		fseek(hFileDat, TempFile->DataPos, SEEK_SET);
		fread(FileData, 0x01, FileLen, hFileDat);
		UnpackFile(FileData, FileLen, &OutBuf, &OutLen);
		free(FileData);	FileData = NULL;
		
		sprintf(OutFileName, "%s%s", Path, TempFile->FileName);
		
		hFile = fopen(OutFileName, "wb");
		if (hFile == NULL)
		{
			printf("Error: Can't open Output File!\n");
			return 0x80;
		}
		
		fwrite(OutBuf, 0x01, OutLen, hFile);
		fclose(hFile);
		
		free(OutBuf);	OutBuf = NULL;
	}
	
	free(OutFileName);
	fclose(hFileDat);
	free(Files);
	
	printf("Done.\n");
	
	return 0x00;
}

void UnpackFile(UINT8* Data, UINT32 DataLen, UINT8** OutBuf, UINT32* OutLen)
{
	UINT32 UncompLen;
	
	memcpy(&UncompLen, &Data[0x00], 0x04);
	
	printf("    Size Compressed: %u, Uncompressed: %u, Ratio: %.2f%%\n", DataLen, UncompLen,
			100.0 * DataLen / UncompLen);
	
	/* extract file without decompressing
	*OutLen = UncompLen;
	*OutBuf = (UINT8*)malloc(*OutLen);
	memcpy(*OutBuf, &Data[0x00], *OutLen);
	*/
	
	// DecompressData calls malloc itself
	*OutLen = 0x00;
	*OutBuf = NULL;
	DecompressData(OutLen, OutBuf, DataLen, Data);
	
	return;
}
